# +
#   ARES/HADES/BORG Package -- ./python/aquila_borg/__init__.py
#   Copyright (C) 2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2020-2021)
#
# +
"""
PyBORG root module
==================

PyBORG is a complete binding of the borg machine to python. The borg
module can be imported either from a normal python interpret or from
the embedded python in hades_python.

Be aware that you cannot import directly any of the native modules. To import
all borg facilities, it is required *and* sufficient to just execute ``import aquila_borg``.
It will provide all the modules such as `aquila_borg.cosmo` or `aquila_borg.samplers`.

If you prefer to have a quiet startup for your scripts, it is possible to disable all output
consoles till you decide to reactivate it later in the execution. You may do so as follow:

.. code:: python

  import os
  os.environ["PYBORG_QUIET"] = "yes"
  import aquila_borg

You should not get any more console message when PyBORG starts.

Many sub-package exist as part of BORG. Among the most important one may find:
    * `aquila_borg.forward`: which deals with forward models
    * `aquila_borg.likelihood`: to express likelihoods
    * `aquila_borg.samplers`: which gives access to available samplers python or C++

These modules includes a `doc` function which provides additional information on the
implementation, the arguments, etc. of the model. As a number of these models are only
available through abstract factories, no python class or functions may be directly
documented.

*Note:* The `borg` module may be installed as an alias to `aquila_borg`. However the package name
is the same as the one used by BorgBackup. Thus there may be clashes between these two softwares.
It is advised to use `aquila_borg` to avoid conflicts.

"""
import sys
import io
import contextlib
import logging
from types import ModuleType
from functools import wraps

try:
    import mpi4py
    import mpi4py.MPI

    MPI4PY_AVAILABLE = True
except:
    MPI4PY_AVAILABLE = False


class _RewriteIO(io.IOBase):
    def __init__(self, original):
        self.original = original
        self.newline = False
        if MPI4PY_AVAILABLE:
            self.comm = mpi4py.MPI.COMM_WORLD
            self.prefix = f"[{self.comm.rank} / {self.comm.size}] "
        else:
            self.prefix = "[0/0]"

    def write(self, b):
        if b[0] == "\n":
            return self.original.write(b)
        return self.original.write(self.prefix + b)

    def writelines(self, lines):
        return self.original.writelines(lines)


@contextlib.contextmanager
def borgConsole(original=sys.stdout):
    """Setup a clean redirection of BORG output, notably in presence of MPI4PY.

    Keyword Arguments:
        original (file-like): Where to send the console to (default: :class:`sys.stdout`)

    """
    with contextlib.redirect_stdout(_RewriteIO(original)):
        yield


with borgConsole():
    from .borg_native import *

from . import dft
from . import utils


class BorgMessageStream(io.IOBase):
    def __init__(self, console, level="std"):
        super().__init__()
        self.console = console
        if level == "std":
            self.printer = console.print_std
        elif level == "warn":
            self.printer = console.print_warning
        elif level == "error":
            self.printer = console.print_error
        else:
            raise ValueError()

    def write(self, b):
        for line in b.split(sep="\n"):
            self.printer(line)

    def writelines(self, lines):
        for line in lines:
            if line[-1] == "\n":
                line = line[:-1]
            self.printer(line)


# This function is to redirect python warning messages to the
# BORG stream for cleanliness
def capture_warnings(logger="py.warnings", level="warn"):
    logging.captureWarnings(True)
    l = logging.getLogger(logger)
    h = logging.StreamHandler(BorgMessageStream(console(), level=level))
    l.addHandler(h)
    return h


@contextlib.contextmanager
def verbose(level: int = 0):
    """Context manager to set a temporary log level for BORG.

    Args:
      * level (int, default 0); Log level to use within the context

    .. code:: python

       with aquila_borg.verbose(3):
          # Some code
          pass
    """
    old = console().getVerboseLevel()
    console().setVerboseLevel(level)
    yield
    console().setVerboseLevel(old)


def getPrintForLevel(level: int):
    if level <= 0:
      return console().print_err
    elif level == 1:
      return console().print_std
    elif level == 2:
      return console().print_warning
    elif level >= 5:
       return console().print_debug

@contextlib.contextmanager
def trace_context(msg: str = None, leave_msg: str = None, level: int = 1):
    """Introduce a context tracing with a python context manager.

    The borg log will be indented.

    Args:
      * msg (str, default None): message to print when entering the context
      * leave_msg (str, default None): message to print when leaving the context
      * level (int, default 1): level at which printing the enter/leave

    Example:

    .. code:: python

       with aquila_borg.trace_context("new stuff"):
          console().print_std("Hello from context")

    """
    if msg is not None:
        getPrintForLevel(level)(msg)
        if leave_msg is None:
            leave_msg = msg
    console()._indent()
    yield
    console()._unindent()
    if leave_msg is not None:
        getPrintForLevel(level)(leave_msg)


def _gen_msg(func, prefix):
    if hasattr(func, "__objclass__"):
        ownclass = func.__objclass__
        if hasattr(ownclass, "__module__"):
            module = ownclass.__module__
            msg = f"{prefix} (module={module} class={objclass}, func={func.__name__})"
        else:
            msg = f"{prefix} (module=UNKNOWN class={objclass}, func={func.__name__})"
    else:
        if hasattr(func, "__module__"):
            module = func.__module__
            msg = f"{prefix} (module={module}, func={func.__name__})"
        else:
            msg = f"{prefix} (module=UNKNOWN, func={func.__name__})"
    return msg


def trace_function(func, msg: str = None):
    if msg is None:
        enter_msg = _gen_msg(func, "Entering")
        leave_msg = _gen_msg(func, "Leaving")
    else:
        enter_msg = f"Entering: {msg}"
        leave_msg = f"Leaving: {msg}"

    @wraps(func)
    def _my_tracer(*args, **kwargs):
        with trace_context(msg=enter_msg, leave_msg=leave_msg):
            func(*args, **kwargs)

    return _my_tracer


def buildDefaultChain(box, cosmo_pars, a_init, model):
    """Build a default BORG chain from a model. This remove the hastle of setting up
    a primordial fluctuation generator, and the cosmological parameters.

    Args:
        box (borg.forward.BoxModel): the specification of the box for the simulation
        cosmo_pars (borg.cosmo.CosmologicalParameters): the cosmology chosen
        a_init (float): the starting scale factor for `model`
        model (borg.forward.BORGForwardModel): the model to be chained with

    Returns:
        borg.forward.BORGForwardModel: a newly built chain for cosmology
    """
    chain = forward.ChainForwardModel(box)
    chain.addModel(forward.models.HermiticEnforcer(box))
    chain.addModel(forward.models.Primordial(box, a_init))
    chain.addModel(forward.models.EisensteinHu(box))
    chain.addModel(model)
    chain.setCosmoParams(cosmo_pars)
    return chain


if EMBEDDED:
    capture_warnings()


# Internal function to retrieve documentation for a model
def _doc_function(name: str) -> None:
    print(forward.models.getDocumentation(name))


# Internal function for injection in a submodule, to document a likelihood.
def _doc_function_lkl(name: str) -> None:
    from pprint import pprint

    print(f"Documentation for '{name}'")
    print(likelihood.getDocumentation(name))
    print()
    print("Further fields required in LikelihoodInfo")
    pprint(likelihood.getInfoRequirement(name))


forward.models.doc = _doc_function
likelihood.doc = _doc_function_lkl


# Create dynamically, from the borg native library, all the forward models
# that are supported with newModel. The docstring is filled and a new submodule
# is populated as forward.model_lib


def _clean_id(s):
    s = "".join(map(lambda x: x if (str.isidentifier(x) or str.isdigit(x)) else "_", s))
    return "M_" + s


def _make_auto_model(model_name):
    def _auto_model(
        box, opts: dict = {}, mpi: object = None, name: str = None
    ) -> forward.BORGForwardModel:
        m = forward.models.newModel(model_name, box, opts=opts, mpi=mpi)
        if name is not None:
            m.setName(name)
        return m

    _auto_model.__name__ = _clean_id(model_name)
    _auto_model.__doc__ = """This is a generic bind for the implicitly registered forward models.
Arguments may be provided throught the 'opts' argument as a key-value pair dictionnary.
The specific model concerned here is '{}'.
{}

   """.format(
        model_name, forward.models.getDocumentation(model_name)
    )
    return _auto_model.__name__, _auto_model


def _make_auto_model_v3(name):
    def _clean(s):
        s = "".join(
            map(lambda x: x if (str.isidentifier(x) or str.isdigit(x)) else "_", s)
        )
        return "M_" + s

    def _auto_model(
        input: modelio.RepresentationDescriptor = None,
        output: modelio.RepresentationDescriptor = None,
        opts: dict = {},
        mpi: object = None,
    ) -> forward.ForwardModel:
        # If there is a dict inside the values, we need to convert it
        # to json
        if any(type(v) is dict for _, v in opts.items()):
            import json

            opts = opts.copy()
            # Convert now
            for k in opts:
                if type(opts[k]) is dict:
                    opts[k] = json.dumps(opts[k])
        return forward.models.newModel_v3(name, input, output, opts=opts, mpi=mpi)

    _auto_model.__name__ = _clean_id(name)
    _auto_model.__doc__ = """This is a generic bind for the implicitly registered forward models.
Arguments may be provided throught the 'opts' argument as a key-value pair
dictionnary.  A difference is that for some arguments that require a json
object (to support dict inside opts) the translation is done on the fly by the
wrapper.

The specific model concerned here is '{}'.
{}

   """.format(
        name, forward.models.getDocumentation(name)
    )
    return _auto_model.__name__, _auto_model


forward.model_lib = ModuleType(
    "model_lib",
    doc="""
Wrapped binding for BORG forward models
=======================================

This is an automatically generated module from the list of available models in the
native side of the BORG library. Models get their name prepended with `M_` to avoid
confusing an identifier for a name and their names gets sanitized for Python language.

.. autosummary::
""",
)
model_lib_names = []

for model_name in forward.models.listModels():
    n, maker = _make_auto_model(model_name)
    model_lib_names.append(n)
    setattr(forward.model_lib, n, maker)

for model_name in forward.models.listModels_v3():
    n, maker = _make_auto_model_v3(model_name)
    model_lib_names.append(n)
    setattr(forward.model_lib, n, maker)

forward.model_lib.__all__ = model_lib_names
forward.model_lib.__doc__ += "\n".join("    {}".format(v) for v in model_lib_names)


def doc(name: str, module: str = None) -> None:
    """Get help information for a BORG abstract or likelihood.

    Args:
      name (str): Name to get information about
      module (str, default None): restrict search to this submodule
            Currently only 'forward' and 'likelihood' are supported.
    """

    if not module is None:
        if module == "forward":
            forward.models.doc(name)
        elif module == "likelihood":
            likelihood.doc(name)
        else:
            raise ValueError("unknown module: " + module)
    else:
        with verbose(-1):
            for m in [forward.models, likelihood]:
                try:
                    m.doc(name)
                except:
                    print("Not found in " + m.__name__)
                    pass


__all__ = [
    "borgConsole",
    "buildDefaultChain",
    "dft",
    "utils",
    "doc",
    "verbose",
    "trace_context",
] + borg_native.__all__

# ARES TAG: authors_num = 1
# ARES TAG: name(0) = Guilhem Lavaux
# ARES TAG: year(0) = 2020
# ARES TAG: email(0) = guilhem.lavaux@iap.fr
